# -----------------------------------------------------------------------------
# Endroid - XMPP Bot - PeriodicPinger plugin
# Copyright 2014, Ensoft Ltd.
# Created by Patrick Stevens
# -----------------------------------------------------------------------------

"""
Periodically selects a user on each server and XMPP-pings them.
It is hoped that this will alleviate problems related to servers dropping our
connections.
"""


import logging
import itertools

import twisted.internet.task

from endroid.pluginmanager import Plugin


class PeriodicPinger(Plugin):
    """
    Wrapper to a Cron job which periodically pings a user on each server.
    This plugin has no XMPP-interface by which the user may interact with it.
    """

    # Class variable to synchronise plugin so that even if we 
    # instantiate more than once, we still only have one Cron job
    task = None

    # Add the hidden attribute so that this plugin doesn't appear in help
    hidden = True

    def endroid_init(self):
        cls = type(self)
        if cls.task is None:
            self.log = logging.getLogger(__name__)
            self.interval = self.vars.get('interval', 10)
            cls.task = twisted.internet.task.LoopingCall(self._ping_all)
            cls.task.start(self.interval)
            self.log.info("Periodic pinger set up for every {} seconds".format(self.interval))

    def _get_one_from_each(self):
        """
        Retrieves a list of 'one user from each server Endroid is connected to'
        :return: list of one person in each server
        """

        def add_resource(user):
            resources = self.usermanagement.resources(user)
            return user + '/' + resources.keys()[0]

        host_users = itertools.groupby(
            sorted(self.usermanagement.available_users(), key=self.usermanagement.get_host),
            key=self.usermanagement.get_host)
        users_to_ping = (next(user) for host, user in host_users)
        uresources_to_ping = (add_resource(user) for user in users_to_ping)
        return uresources_to_ping

    def _ping_all(self):
        """Pings a user on each server Endroid knows about."""

        users_to_ping = set(self._get_one_from_each())
        logging.debug('PeriodicPinger: Pinging {}...'.format(users_to_ping))
        for user in users_to_ping:
            self.usermanagement.ping(user)

